@@ -25,6 +25,7 @@ import fasteners
 import networkx as nx
 from oslo_utils import excutils
 from oslo_utils import strutils
+from oslo_utils import timeutils
 import six
 
 from taskflow.engines.action_engine import builder
@@ -116,6 +117,15 @@ class ActionEngine(base.Engine):
     |                      | has a separate retry  |      |            |
     |                      | behavior).            |      |            |
     +----------------------+-----------------------+------+------------+
+    | ``never_resolve``    | When true, instead    | bool | ``False``  |
+    |                      | of reverting          |      |            |
+    |                      | and trying to resolve |      |            |
+    |                      | a atom failure the    |      |            |
+    |                      | engine will skip      |      |            |
+    |                      | reverting and abort   |      |            |
+    |                      | instead of reverting  |      |            |
+    |                      | and/or retrying.      |      |            |
+    +----------------------+-----------------------+------+------------+
     | ``inject_transient`` | When true, values     | bool | ``True``   |
     |                      | that are local to     |      |            |
     |                      | each atoms scope      |      |            |
@@ -161,7 +171,6 @@ class ActionEngine(base.Engine):
         self._compilation = None
         self._compiler = compiler.PatternCompiler(flow)
         self._lock = threading.RLock()
-        self._state_lock = threading.RLock()
         self._storage_ensured = False
         self._validated = False
         # Retries are not *currently* executed out of the engines process
@@ -169,6 +178,9 @@ class ActionEngine(base.Engine):
         self._retry_executor = executor.SerialRetryExecutor()
         self._inject_transient = strutils.bool_from_string(
             self._options.get('inject_transient', True))
+        self._gather_statistics = strutils.bool_from_string(
+            self._options.get('gather_statistics', True))
+        self._statistics = {}
 
     @_pre_check(check_compiled=True,
                 # NOTE(harlowja): We can alter the state of the
@@ -180,6 +192,10 @@ class ActionEngine(base.Engine):
     def suspend(self):
         self._change_state(states.SUSPENDING)
 
+    @property
+    def statistics(self):
+        return self._statistics
+
     @property
     def compilation(self):
         """The compilation result.
@@ -216,12 +232,18 @@ class ActionEngine(base.Engine):
                                backend=self._backend,
                                scope_fetcher=_scope_fetcher)
 
-    def run(self):
+    def run(self, timeout=None):
+        """Runs the engine (or die trying).
+
+        :param timeout: timeout to wait for any atoms to complete (this timeout
+            will be used during the waiting period that occurs when
+            unfinished atoms are being waited on).
+        """
         with fasteners.try_lock(self._lock) as was_locked:
             if not was_locked:
                 raise exc.ExecutionFailure("Engine currently locked, please"
                                            " try again later")
-            for _state in self.run_iter():
+            for _state in self.run_iter(timeout=timeout):
                 pass
 
     def run_iter(self, timeout=None):
@@ -255,9 +277,17 @@ class ActionEngine(base.Engine):
             maxlen=max(1, self.MAX_MACHINE_STATES_RETAINED))
         with _start_stop(self._task_executor, self._retry_executor):
             self._change_state(states.RUNNING)
+            if self._gather_statistics:
+                self._statistics.clear()
+                w = timeutils.StopWatch()
+                w.start()
+            else:
+                w = None
             try:
                 closed = False
-                machine, memory = self._runtime.builder.build(timeout=timeout)
+                machine, memory = self._runtime.builder.build(
+                    self._statistics, timeout=timeout,
+                    gather_statistics=self._gather_statistics)
                 r = runners.FiniteRunner(machine)
                 for transition in r.run_iter(builder.START):
                     last_transitions.append(transition)
@@ -277,6 +307,13 @@ class ActionEngine(base.Engine):
                         # shop...
                         closed = True
                         self.suspend()
+                    except Exception:
+                        # Capture the failure, and ensure that the
+                        # machine will notice that something externally
+                        # has sent an exception in and that it should
+                        # finish up and reraise.
+                        memory.failures.append(failure.Failure())
+                        closed = True
                     else:
                         if try_suspend:
                             self.suspend()
@@ -300,6 +337,10 @@ class ActionEngine(base.Engine):
                                 six.itervalues(failures),
                                 six.itervalues(more_failures))
                             failure.Failure.reraise_if_any(fails)
+            finally:
+                if w is not None:
+                    w.stop()
+                    self._statistics['active_for'] = w.elapsed()
 
     @staticmethod
     def _check_compilation(compilation):
@@ -320,18 +361,15 @@ class ActionEngine(base.Engine):
         return compilation
 
     def _change_state(self, state):
-        with self._state_lock:
-            old_state = self.storage.get_flow_state()
-            if not states.check_flow_transition(old_state, state):
-                return
-            self.storage.set_flow_state(state)
-        details = {
-            'engine': self,
-            'flow_name': self.storage.flow_name,
-            'flow_uuid': self.storage.flow_uuid,
-            'old_state': old_state,
-        }
-        self.notifier.notify(state, details)
+        moved, old_state = self.storage.change_flow_state(state)
+        if moved:
+            details = {
+                'engine': self,
+                'flow_name': self.storage.flow_name,
+                'flow_uuid': self.storage.flow_uuid,
+                'old_state': old_state,
+            }
+            self.notifier.notify(state, details)
 
     def _ensure_storage(self):
         """Ensure all contained atoms exist in the storage unit."""
@@ -364,8 +402,12 @@ class ActionEngine(base.Engine):
         last_node = None
         missing_nodes = 0
         for atom in self._runtime.analyzer.iterate_nodes(compiler.ATOMS):
-            atom_missing = self.storage.fetch_unsatisfied_args(
+            exec_missing = self.storage.fetch_unsatisfied_args(
                 atom.name, atom.rebind, optional_args=atom.optional)
+            revert_missing = self.storage.fetch_unsatisfied_args(
+                atom.name, atom.revert_rebind,
+                optional_args=atom.revert_optional)
+            atom_missing = exec_missing.union(revert_missing)
             if atom_missing:
                 cause = exc.MissingDependencies(atom,
                                                 sorted(atom_missing),
@@ -477,6 +519,12 @@ String (case insensitive)    Executor used
 ``thread``                   :class:`~.executor.ParallelThreadTaskExecutor`
 ``threaded``                 :class:`~.executor.ParallelThreadTaskExecutor`
 ``threads``                  :class:`~.executor.ParallelThreadTaskExecutor`
+``greenthread``              :class:`~.executor.ParallelThreadTaskExecutor`
+                              (greened version)
+``greedthreaded``            :class:`~.executor.ParallelThreadTaskExecutor`
+                              (greened version)
+``greenthreads``             :class:`~.executor.ParallelThreadTaskExecutor`
+                              (greened version)
 ===========================  ===============================================
 
     * ``max_workers``: a integer that will affect the number of parallel
@@ -520,6 +568,9 @@ String (case insensitive)    Executor used
                            executor.ParallelProcessTaskExecutor),
         _ExecutorTextMatch(frozenset(['thread', 'threads', 'threaded']),
                            executor.ParallelThreadTaskExecutor),
+        _ExecutorTextMatch(frozenset(['greenthread', 'greenthreads',
+                                      'greenthreaded']),
+                           executor.ParallelGreenThreadTaskExecutor),
     ]
 
     # Used when no executor is provided (either a string or object)...
