@@ -17,6 +17,7 @@
 import contextlib
 import string
 import threading
+import time
 
 import redis
 import six
@@ -117,6 +118,19 @@ class AddOne(task.Task):
         return source + 1
 
 
+class GiveBackRevert(task.Task):
+
+    def execute(self, value):
+        return value + 1
+
+    def revert(self, *args, **kwargs):
+        result = kwargs.get('result')
+        # If this somehow fails, timeout, or other don't send back a
+        # valid result...
+        if isinstance(result, six.integer_types):
+            return result + 1
+
+
 class FakeTask(object):
 
     def execute(self, **kwargs):
@@ -319,6 +333,19 @@ class NeverRunningTask(task.Task):
         assert False, 'This method should not be called'
 
 
+class TaskRevertExtraArgs(task.Task):
+    def execute(self, **kwargs):
+        raise exceptions.ExecutionFailure("We want to force a revert here")
+
+    def revert(self, revert_arg, flow_failures, result, **kwargs):
+        pass
+
+
+class SleepTask(task.Task):
+    def execute(self, duration, **kwargs):
+        time.sleep(duration)
+
+
 class EngineTestBase(object):
     def setUp(self):
         super(EngineTestBase, self).setUp()
